package com.sun.xml.internal.ws.util;

import java.util.Iterator;
import java.util.Collection;
import java.lang.reflect.Field;
import java.util.List;
import java.lang.reflect.Modifier;
import javax.annotation.Resource;
import java.util.ArrayList;
import java.security.AccessController;
import java.lang.reflect.InvocationTargetException;
import java.security.PrivilegedAction;
import java.lang.reflect.Method;
import javax.xml.ws.WebServiceException;
import java.util.concurrent.Callable;

public abstract class InjectionPlan<T, R>
{
    public InjectionPlan() {
    }
    
    public abstract void inject(T p0, R p1);
    
    public void inject(T instance, Callable<R> resource) {
        try {
            this.inject(instance, resource.call());
        }
        catch (Exception e) {
            throw new WebServiceException(e);
        }
    }
    
    private static void invokeMethod(final Method method, final Object instance, final Object... args) {
        if (method == null) {
            return;
        }
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                try {
                    if (!method.isAccessible()) {
                        method.setAccessible(true);
                    }
                    method.invoke(instance, args);
                }
                catch (IllegalAccessException e) {
                    throw new WebServiceException(e);
                }
                catch (InvocationTargetException e2) {
                    throw new WebServiceException(e2);
                }
                return null;
            }
        });
        return;
    }
    
    public static <T, R> InjectionPlan<T, R> buildInjectionPlan(Class<? extends T> clazz, Class<R> resourceType, boolean isStatic) {
        List<InjectionPlan<T, R>> plan = new ArrayList();
        for (Class<?> cl = clazz; cl != Object.class; cl = cl.getSuperclass()) {
            for (Field field : cl.getDeclaredFields()) {
                Resource resource = (Resource)field.getAnnotation(Resource.class);
                if (resource != null && isInjectionPoint(resource, field.getType(), "Incorrect type for field" + field.getName(), resourceType)) {
                    if (isStatic && !Modifier.isStatic(field.getModifiers())) {
                        throw new WebServiceException("Static resource " + resourceType + " cannot be injected to non-static " + field);
                    }
                    plan.add(new FieldInjectionPlan(field));
                }
            }
        }
        for (Class<?> cl = clazz; cl != Object.class; cl = cl.getSuperclass()) {
            for (Method method : cl.getDeclaredMethods()) {
                Resource resource = (Resource)method.getAnnotation(Resource.class);
                if (resource != null) {
                    Class[] paramTypes = method.getParameterTypes();
                    if (paramTypes.length != 1) {
                        throw new WebServiceException("Incorrect no of arguments for method " + method);
                    }
                    if (isInjectionPoint(resource, paramTypes[0], "Incorrect argument types for method" + method.getName(), resourceType)) {
                        if (isStatic && !Modifier.isStatic(method.getModifiers())) {
                            throw new WebServiceException("Static resource " + resourceType + " cannot be injected to non-static " + method);
                        }
                        plan.add(new MethodInjectionPlan(method));
                    }
                }
            }
        }
        return new InjectionPlan.Compositor(plan);
    }
    
    private static boolean isInjectionPoint(Resource resource, Class fieldType, String errorMessage, Class resourceType) {
        Class<Object> t = resource.type();
        if (t.equals(Object.class)) {
            return fieldType.equals(resourceType);
        }
        if (t.equals(resourceType)) {
            if (fieldType.isAssignableFrom(resourceType)) {
                return true;
            }
            throw new WebServiceException(errorMessage);
        }
        else {
            return false;
        }
    }
    
    static void access$100(Method x0, Object x1, Object[] x2) {
        invokeMethod(x0, x1, x2);
    }
    
    public static class FieldInjectionPlan<T, R> extends InjectionPlan<T, R>
    {
        private final Field field;
        
        public FieldInjectionPlan(Field field) {
            this.field = field;
        }
        
        @Override
        public void inject(T instance, R resource) {
            AccessController.doPrivileged(new PrivilegedAction<Object>() {
                @Override
                public Object run() {
                    try {
                        if (!InjectionPlan.FieldInjectionPlan.this.field.isAccessible()) {
                            InjectionPlan.FieldInjectionPlan.this.field.setAccessible(true);
                        }
                        InjectionPlan.FieldInjectionPlan.this.field.set(instance, resource);
                        return null;
                    }
                    catch (IllegalAccessException e) {
                        throw new WebServiceException(e);
                    }
                }
            });
        }
    }
    
    public static class MethodInjectionPlan<T, R> extends InjectionPlan<T, R>
    {
        private final Method method;
        
        public MethodInjectionPlan(Method method) {
            this.method = method;
        }
        
        @Override
        public void inject(T instance, R resource) {
            InjectionPlan.access$100(this.method, instance, new Object[] { resource });
        }
    }
    
    private static class Compositor<T, R> extends InjectionPlan<T, R>
    {
        private final Collection<InjectionPlan<T, R>> children;
        
        public Compositor(Collection<InjectionPlan<T, R>> children) {
            this.children = children;
        }
        
        @Override
        public void inject(T instance, R res) {
            for (InjectionPlan<T, R> plan : this.children) {
                plan.inject(instance, res);
            }
        }
        
        @Override
        public void inject(T instance, Callable<R> resource) {
            if (!this.children.isEmpty()) {
                super.inject(instance, resource);
            }
        }
    }
}
